import pydantic

from .base import Base, LowercaseStrEnum


class RadialGridType(LowercaseStrEnum):
    """What sort of radial grid (only one option for now)"""

    KRACK_KOSTER = "krack_koster"
    LMG = "lmg"


class GridSettings(Base):
    radial_grid_type: RadialGridType = RadialGridType.LMG
    angular_num_points: pydantic.PositiveInt = 434

    weight_cutoff: pydantic.PositiveFloat = 1e-11

    # for LMG
    radial_precision: pydantic.PositiveFloat = 1e-11

    # for other schemes, like KK
    radial_num_points: pydantic.PositiveInt = 75

    # pruning?
    prune: bool = True
    min_angular_points: pydantic.PositiveInt = 50



from pydantic import PositiveFloat, PositiveInt

from .base import Base, LowercaseStrEnum


class ImplicitSolventModel(LowercaseStrEnum):
    CPCM = "cpcm"
    ALPB = "alpb"
    COSMO = "cosmo"
    NONE = "none"


class SolventSettings(Base):
    model: ImplicitSolventModel = ImplicitSolventModel.NONE
    epsilon: float = 78.36

    grid_points_per_atom: PositiveInt = 170
    vdw_scale: PositiveFloat = 1.2
    weight_cutoff: PositiveFloat = 1e-8



from .base import Base, LowercaseStrEnum


class ComputeType(LowercaseStrEnum):
    CPU = "cpu"
    GPU = "gpu"
    AUTO = "auto"


class ComputeSettings(Base):
    requested_compute_type: ComputeType = ComputeType.CPU
    compute_type_used: ComputeType | None = None



from typing import Literal

from .base import LowercaseStrEnum


class Method(LowercaseStrEnum):
    HARTREE_FOCK = "hf"
    HF3C = "hf_3c"

    PBE = "pbe"
    B973C = "b97_3c"
    B97D3BJ = "b97_d3bj"
    R2SCAN = "r2scan"
    R2SCAN3C = "r2scan_3c"
    TPSS = "tpss"
    M06L = "m06l"

    PBE0 = "pbe0"
    B3LYP = "b3lyp"
    TPSSH = "tpssh"
    M06 = "m06"
    M062X = "m062x"

    CAMB3LYP = "camb3lyp"
    WB97XD3 = "wb97x_d3"
    WB97XV = "wb97x_v"
    WB97MV = "wb97m_v"
    WB97MD3BJ = "wb97m_d3bj"
    WB97X3C = "wb97x_3c"

    DSDBLYPD3BJ = "dsd_blyp_d3bj"

    AIMNET2_WB97MD3 = "aimnet2_wb97md3"
    MACE_MP_0 = "mace_mp_0"
    MACE_MP_0B2_L = "mace_mp_0b2_l"
    OCP24_S = "ocp24_s"
    OCP24_L = "ocp24_l"
    ORB_V2 = "orb_v2"

    GFN_FF = "gfn_ff"
    GFN0_XTB = "gfn0_xtb"
    GFN1_XTB = "gfn1_xtb"
    GFN2_XTB = "gfn2_xtb"

    # this was going to be removed, but Jonathon wrote such a nice basis set test... it's off the front end.
    BP86 = "bp86"

    OFF_SAGE_2_2_1 = "off_sage_2_2_1"

    RM1 = "rm1"


PrepackagedNNPMethod = Literal[Method.AIMNET2_WB97MD3, Method.OCP24_S, Method.OCP24_L, Method.RM1]
PREPACKAGED_NNP_METHODS = [Method.AIMNET2_WB97MD3, Method.OCP24_S, Method.OCP24_L, Method.RM1]

CorrectableNNPMethod = Literal[Method.MACE_MP_0B2_L, Method.ORB_V2]
CORRECTABLE_NNP_METHODS = [Method.MACE_MP_0B2_L, Method.ORB_V2]

NNPMethod = PrepackagedNNPMethod | CorrectableNNPMethod
NNP_METHODS = [*PREPACKAGED_NNP_METHODS, *CORRECTABLE_NNP_METHODS]

XTBMethod = Literal[Method.GFN_FF, Method.GFN0_XTB, Method.GFN1_XTB, Method.GFN2_XTB]
XTB_METHODS = [Method.GFN_FF, Method.GFN0_XTB, Method.GFN1_XTB, Method.GFN2_XTB]

CompositeMethod = Literal[Method.HF3C, Method.B973C, Method.R2SCAN3C, Method.WB97X3C]
COMPOSITE_METHODS = [Method.HF3C, Method.B973C, Method.R2SCAN3C, Method.WB97X3C]

FFMethod = Literal[Method.OFF_SAGE_2_2_1]
FF_METHODS = [Method.OFF_SAGE_2_2_1]

PrepackagedMethod = XTBMethod | CompositeMethod | PrepackagedNNPMethod | FFMethod
PREPACKAGED_METHODS = [*XTB_METHODS, *COMPOSITE_METHODS, *PREPACKAGED_NNP_METHODS, *FF_METHODS]

MethodWithCorrection = Literal[Method.WB97XD3, Method.WB97XV, Method.WB97MV, Method.WB97MD3BJ, Method.DSDBLYPD3BJ]
METHODS_WITH_CORRECTION = [Method.WB97XD3, Method.WB97XV, Method.WB97MV, Method.WB97MD3BJ, Method.DSDBLYPD3BJ, Method.B97D3BJ]



from typing import Any, Optional, Self, TypeVar

from pydantic import computed_field, field_validator, model_validator

from .base import Base, UniqueList
from .basis_set import BasisSet
from .compute_settings import ComputeSettings
from .correction import Correction
from .method import CORRECTABLE_NNP_METHODS, METHODS_WITH_CORRECTION, PREPACKAGED_METHODS, Method
from .mode import Mode
from .opt_settings import OptimizationSettings
from .scf_settings import SCFSettings
from .solvent import SolventSettings
from .task import Task
from .thermochem_settings import ThermochemistrySettings

_T = TypeVar("_T")


class Settings(Base):
    mode: Mode = Mode.AUTO

    method: Method = Method.HARTREE_FOCK
    basis_set: Optional[BasisSet] = None
    tasks: UniqueList[Task] = [Task.ENERGY, Task.CHARGE, Task.DIPOLE]
    corrections: UniqueList[Correction] = []

    solvent_settings: Optional[SolventSettings] = None

    # scf/opt settings will be set automatically based on mode, but can be overridden manually
    scf_settings: SCFSettings = SCFSettings()
    opt_settings: OptimizationSettings = OptimizationSettings()
    thermochem_settings: ThermochemistrySettings = ThermochemistrySettings()
    compute_settings: ComputeSettings = ComputeSettings()

    # mypy has this dead wrong (https://docs.pydantic.dev/2.0/usage/computed_fields/)
    # Python 3.12 narrows the reason for the ignore to prop-decorator
    @computed_field  # type: ignore[misc, prop-decorator, unused-ignore]
    @property
    def level_of_theory(self) -> str:
        corrections = list(filter(lambda x: x not in (None, ""), self.corrections))

        if self.method in CORRECTABLE_NNP_METHODS:
            method = self.method.value if not corrections else f"{self.method.value}-{'-'.join(c.value for c in corrections)}"
        elif self.method in PREPACKAGED_METHODS or self.basis_set is None:
            method = self.method.value
        elif self.method in METHODS_WITH_CORRECTION or not corrections:
            method = f"{self.method.value}/{self.basis_set.name.lower()}"
        else:
            method = f"{self.method.value}-{'-'.join(c.value for c in corrections)}/{self.basis_set.name.lower()}"

        if self.solvent_settings is not None:
            method += f"/{self.solvent_settings.model.value}({self.solvent_settings.solvent.value})"

        return method

    @field_validator("mode")
    @classmethod
    def set_mode_auto(cls, mode: Mode) -> Mode:
        """Set the mode to RAPID if AUTO is selected."""
        if mode == Mode.AUTO:
            return Mode.RAPID

        return mode

    @model_validator(mode="after")
    def validate_and_build(self) -> Self:
        if self.mode == Mode.AUTO:
            self.mode = Mode.RAPID

        self.scf_settings = _assign_scf_settings_by_mode(self.mode, self.scf_settings)
        self.opt_settings = _assign_opt_settings_by_mode(self.mode, self.opt_settings)

        return self

    def model_post_init(self, __context: Any) -> None:
        # figure out `optimize_ts`
        if Task.OPTIMIZE_TS in self.tasks:
            self.tasks.pop(self.tasks.index(Task.OPTIMIZE_TS))
            self.tasks.append(Task.OPTIMIZE)
            self.opt_settings.transition_state = True

        # composite methods have their own basis sets, so overwrite user stuff
        if self.method == Method.HF3C:
            self.basis_set = BasisSet(name="minix")
        elif self.method == Method.B973C:
            self.basis_set = BasisSet(name="def2-mTZVP")
        elif self.method == Method.R2SCAN3C:
            self.basis_set = BasisSet(name="def2-mTZVPP")
        elif self.method == Method.WB97X3C:
            self.basis_set = BasisSet(name="vDZP")

    @field_validator("basis_set", mode="before")
    @classmethod
    def parse_basis_set(cls, v: Any) -> BasisSet | dict[str, Any] | None:
        """Turn a string into a ``BasisSet`` object. (This is a little crude.)"""
        if isinstance(v, BasisSet):
            return None if v.name is None else v
        elif isinstance(v, dict):
            return None if v.get("name") is None else v
        elif isinstance(v, str):
            if len(v):
                return BasisSet(name=v)
            # "" is basically None, let's be real here...
            return None
        elif v is None:
            return None
        else:
            raise ValueError(f"invalid value ``{v}`` for ``basis_set``")

    @field_validator("corrections", mode="before")
    @classmethod
    def remove_empty_string(cls, v: list[_T]) -> list[_T]:
        """Remove empty string values."""
        return [c for c in v if c] if v is not None else v


def _assign_scf_settings_by_mode(mode: Mode, scf_settings: SCFSettings) -> SCFSettings:
    """
    Assign SCF settings based on the mode.

    Values based off of the following sources:
    QChem:
        - https://manual.q-chem.com/5.2/Ch4.S3.SS2.html
        - https://manual.q-chem.com/5.2/Ch4.S5.SS2.html

    Gaussian:
        - https://gaussian.com/integral/
        - https://gaussian.com/overlay5/

    Orca:
        - manual 4.2.1, §9.6.1 and §9.7.3

    Psi4:
        - https://psicode.org/psi4manual/master/autodir_options_c/module__scf.html
        - https://psicode.org/psi4manual/master/autodoc_glossary_options_c.html

    TeraChem:
        - Manual, it's easy to locate everything.

    The below values are my best attempt at homogenizing various sources.
    In general, eri_threshold should be 3 OOM lower than SCF convergence.
    """
    if mode == Mode.MANUAL:
        return scf_settings

    match mode:
        case Mode.RECKLESS:
            scf_settings.energy_threshold = 1e-5
            scf_settings.rms_error_threshold = 1e-7
            scf_settings.max_error_threshold = 1e-5
            scf_settings.rebuild_frequency = 100
            scf_settings.int_settings.eri_threshold = 1e-8
            scf_settings.int_settings.csam_multiplier = 3.0
            scf_settings.int_settings.pair_overlap_threshold = 1e-8
        case Mode.RAPID | Mode.CAREFUL:
            scf_settings.energy_threshold = 1e-6
            scf_settings.rms_error_threshold = 1e-9
            scf_settings.max_error_threshold = 1e-7
            scf_settings.rebuild_frequency = 10
            scf_settings.int_settings.eri_threshold = 1e-10
            scf_settings.int_settings.csam_multiplier = 1.0
            scf_settings.int_settings.pair_overlap_threshold = 1e-10
        case Mode.METICULOUS:
            scf_settings.energy_threshold = 1e-8
            scf_settings.rms_error_threshold = 1e-9
            scf_settings.max_error_threshold = 1e-7
            scf_settings.rebuild_frequency = 5
            scf_settings.int_settings.eri_threshold = 1e-12
            scf_settings.int_settings.csam_multiplier = 1.0
            scf_settings.int_settings.pair_overlap_threshold = 1e-12
        case Mode.DEBUG:
            scf_settings.energy_threshold = 1e-9
            scf_settings.rms_error_threshold = 1e-10
            scf_settings.max_error_threshold = 1e-9
            scf_settings.rebuild_frequency = 1
            scf_settings.int_settings.eri_threshold = 1e-14
            scf_settings.int_settings.csam_multiplier = 1e10  # in other words, disable CSAM
            scf_settings.int_settings.pair_overlap_threshold = 1e-14
        case _:
            raise ValueError(f"Unknown mode ``{mode.value}``!")

    return scf_settings


def _assign_opt_settings_by_mode(mode: Mode, opt_settings: OptimizationSettings) -> OptimizationSettings:
    """
    Assign optimization settings based on the mode.

    Constraints lead to a lot of noise, so we need to loosen the thresholds.

    cf. DLFIND manual, and https://www.cup.uni-muenchen.de/ch/compchem/geom/basic.html
    and the discussion at https://geometric.readthedocs.io/en/latest/how-it-works.html
    in periodic systems, "normal" is 0.05 eV/Å ~= 2e-3 Hartree/Å, and "careful" is 0.01 ~= 4e-4

    Note: thresholds here are in units of Hartree/Å, not Hartree/Bohr as listed in many places.
    """
    opt_settings.energy_threshold = 1e-6
    match mode:
        case Mode.RECKLESS:
            opt_settings.energy_threshold = 2e-5
            opt_settings.max_gradient_threshold = 7e-3
            opt_settings.rms_gradient_threshold = 6e-3
        case Mode.RAPID:
            opt_settings.energy_threshold = 5e-5
            opt_settings.max_gradient_threshold = 5e-3
            opt_settings.rms_gradient_threshold = 3.5e-3
        case Mode.CAREFUL:
            opt_settings.max_gradient_threshold = 9e-4
            opt_settings.rms_gradient_threshold = 6e-4
        case Mode.METICULOUS:
            opt_settings.max_gradient_threshold = 3e-5
            opt_settings.rms_gradient_threshold = 2e-5
        case Mode.DEBUG:
            opt_settings.max_gradient_threshold = 4e-6
            opt_settings.rms_gradient_threshold = 2e-6
        case _:
            raise ValueError(f"Unknown mode ``{mode.value}``!")

    return opt_settings



from typing import Sequence

from pydantic import PositiveFloat, PositiveInt

from .base import Base
from .constraint import Constraint


class OptimizationSettings(Base):
    max_steps: PositiveInt = 250
    transition_state: bool = False

    # when are we converged? (Hartree and Hartree/Å)
    max_gradient_threshold: PositiveFloat = 7e-4
    rms_gradient_threshold: PositiveFloat = 6e-4
    energy_threshold: PositiveFloat = 1e-6

    # for periodic systems only
    optimize_cell: bool = False

    constraints: Sequence[Constraint] = tuple()



from .base import LowercaseStrEnum


class Correction(LowercaseStrEnum):
    """Various post hoc corrections."""

    # Grimme's D3 dispersion correction, with Becke–Johnson damping
    D3BJ = "d3bj"

    # Grimme's D3 dispersion correction, *without* Becke–Johnson damping
    D3 = "d3"

    # Grimme's D4 dispersion correction
    D4 = "d4"

    # Grimme's geometric counterpoise correction
    GCP = "gcp"



import pydantic

from .base import Base, LowercaseStrEnum


class ERIStrategy(LowercaseStrEnum):
    # direct SCF, as per Almlof/Faegri/Korsell
    DIRECT = "direct"

    # Raffinetti-style supermatrix
    SUPERMATRIX = "supermatrix"

    # let the software choose between direct and supermatrix based on size
    AUTO = "auto"

    # resolution of the identity for J and K
    RIJK = "rijk"


class IntSettings(Base):
    strategy: ERIStrategy = ERIStrategy.AUTO

    # these will get overwritten by ``mode`` anyway, for the most part
    eri_threshold: pydantic.PositiveFloat = 1e-9
    csam_multiplier: pydantic.PositiveFloat = pydantic.Field(default=1, ge=1)
    pair_overlap_threshold: pydantic.PositiveFloat = 1e-10

    @property
    def resolution_of_the_identity(self) -> bool:
        """Abstracting in case we add RIJCOSX, etc later"""
        return self.strategy == ERIStrategy.RIJK



from typing import Callable, Iterable, TypeAlias

UUID: TypeAlias = str

Vector3D: TypeAlias = tuple[float, float, float]
Vector3DPerAtom: TypeAlias = list[Vector3D]

FloatPerAtom: TypeAlias = list[float]

Matrix3x3: TypeAlias = tuple[Vector3D, Vector3D, Vector3D]


def round_list(round_to: int = 6) -> Callable[[Iterable[float]], list[float]]:
    """Create a validator that rounds each float in a list to a given number of decimal places."""

    def rounder(values: Iterable[float]) -> list[float]:
        return [round(value, round_to) for value in values]

    return rounder


def round_vector3d(round_to: int = 6) -> Callable[[Vector3D], Vector3D]:
    """Create a validator that rounds each component of a Vector3D to a given number of decimal places."""

    def rounder(vector: Vector3D) -> Vector3D:
        return (round(vector[0], round_to), round(vector[1], round_to), round(vector[2], round_to))

    return rounder


def round_optional_vector3d(round_to: int = 6) -> Callable[[Vector3D | None], Vector3D | None]:
    """Create a validator that rounds each component of a Vector3D to a given number of decimal places, handling None."""

    def rounder(vector: Vector3D | None) -> Vector3D | None:
        if vector is None:
            return None
        return (round(vector[0], round_to), round(vector[1], round_to), round(vector[2], round_to))

    return rounder


def round_vector3d_per_atom(round_to: int = 6) -> Callable[[Vector3DPerAtom], Vector3DPerAtom]:
    """Create a validator that rounds each vector in Vector3DPerAtom to a given number of decimal places."""
    vector_rounder = round_vector3d(round_to)

    def rounder(vectors: Vector3DPerAtom) -> Vector3DPerAtom:
        return [vector_rounder(v) for v in vectors]

    return rounder


def round_optional_vector3d_per_atom(round_to: int = 6) -> Callable[[Vector3DPerAtom | None], Vector3DPerAtom | None]:
    """Create a validator that rounds each vector in Vector3DPerAtom to a given number of decimal places, handling None."""
    vector_rounder = round_vector3d(round_to)

    def rounder(vectors: Vector3DPerAtom | None) -> Vector3DPerAtom | None:
        if vectors is None:
            return None
        return [vector_rounder(v) for v in vectors]

    return rounder


def round_matrix3x3(round_to: int = 6) -> Callable[[Matrix3x3], Matrix3x3]:
    """Create a validator that rounds each vector in a Matrix3x3 to a given number of decimal places."""

    # Use the round_vector3d function to round each Vector3D in the Matrix3x3
    vector_rounder = round_vector3d(round_to)

    def rounder(matrix: Matrix3x3) -> Matrix3x3:
        return (vector_rounder(matrix[0]), vector_rounder(matrix[1]), vector_rounder(matrix[2]))

    return rounder


def round_optional_matrix3x3(round_to: int = 3) -> Callable[[Matrix3x3 | None], Matrix3x3 | None]:
    """Create a validator that rounds each vector in an Optional Matrix3x3 to a given number of decimal places."""

    # Use the round_vector3d function to round each Vector3D in the Matrix3x3
    vector_rounder = round_vector3d(round_to)

    def rounder(matrix: Matrix3x3 | None) -> Matrix3x3 | None:
        if matrix is None:
            return None
        return (vector_rounder(matrix[0]), vector_rounder(matrix[1]), vector_rounder(matrix[2]))

    return rounder


def round_optional_float_per_atom(round_to: int = 6) -> Callable[[FloatPerAtom | None], FloatPerAtom | None]:
    """Create a validator that rounds each float in FloatPerAtom to a given number of decimal places, handling None."""

    def rounder(values: FloatPerAtom | None) -> FloatPerAtom | None:
        if values is None:
            return None
        return [round(value, round_to) for value in values]

    return rounder


def round_float_per_atom(round_to: int = 6) -> Callable[[FloatPerAtom], FloatPerAtom]:
    """Create a validator that rounds each float in FloatPerAtom to a given number of decimal places, handling None."""

    def rounder(values: FloatPerAtom) -> FloatPerAtom:
        return [round(value, round_to) for value in values]

    return rounder
